#!/usr/bin/perl

use strict;
use warnings;
use Test::More;
use FDN::Common::Remember;

my $nbtest;

# On fait de la concurrence, ne pas numeroter les tests
Test::More->builder->use_numbers(0);

for my $name ( 'rem1', 'rem2' ) {
	for my $ext ( 'lock', 'latest', 'data' ) {
		my $file = './t/'.$name.'.'.$ext;
		if ( -e $file ) {
			unlink $file || die "Failed to unlink $file";
		}
	}
}


setup_remember('./t','lock');
is ( $FDN::Common::Remember::REP, './t', 'setup_remember' );
is ( $FDN::Common::Remember::MODE, 'lock', 'setup_remember' );
BEGIN { $nbtest += 2; }

my $job;

$job = 'rem1';

my ( $res, $tmp );
$res = process_start($job);
ok ( $res, 'process_start' );
ok ( -e "./t/$job.lock", 'process_start' );
ok ( ! -e "./t/$job.nolock", 'process_start' );
ok ( ! -e "./t/$job.latest", 'process_start' );
ok ( ! -e "./t/$job.data ", 'process_start' );
BEGIN { $nbtest += 5; }

$res = process_get_latest($job);
is ( $res, undef, 'process_get_latest' );
BEGIN { $nbtest += 1; }

$res = process_stop($job);
ok ( $res, 'process_stop' );
BEGIN { $nbtest += 1; }

my $nb_sub1;
subtest 'concurrent 1' => sub {
	plan tests => $nb_sub1;
	Test::More->builder->use_numbers(0);
	if ( fork() ) {
		$res = process_start($job);
		ok ( $res, 'process_start' );
		$res = process_get_latest($job);
		is ( ref($res), 'HASH', 'process_get_latest' );
		ok ( defined $res->{'start'}, 'process_get_latest' );
		ok ( defined $res->{'end'} , 'process_get_latest' );
		ok ( defined $res->{'hostname'}, 'process_get_latest' );
		ok ( defined $res->{'pid'}, 'process_get_latest' );
		ok ( time() - $res->{'start'} < 2, 'process_get_latest' );
		ok ( $res->{'pid'} == $$, 'process_get_latest' );
		BEGIN { $nb_sub1 += 8; };
		sleep(2);
		$res = process_stop($job);
		ok ( $res, 'process_stop' );
		BEGIN { $nb_sub1 += 1; }
	} else {
		sleep(1);
		$res = process_start($job);
		ok ( ! $res, 'process_start');
		$res = process_get_latest($job);
		is ( ref($res), 'HASH', 'process_get_latest');
		eval { $res = process_abort($job); };
		ok ( $@, 'process_abort' );
		BEGIN { $nb_sub1 += 3; }
		# Je suis le fils, je meurs.
		if ( ! Test::More->builder->is_passing() ) {
			exit(1); # Echec
		} else {
			exit(0); # Success
		}
	}
	my $wait = wait();
	Test::More->builder->current_test($nb_sub1 - 1);
	is ( $?, 0, 'fin_du_fils' );
	BEGIN { $nb_sub1 += 1; }
};
BEGIN { $nbtest += 1; }

sleep(4);

my $nb_sub2;
subtest 'concurrent 2' => sub {
	plan tests => $nb_sub2;
	Test::More->builder->use_numbers(0);
	if ( fork() ) {
		$res = process_start($job);
		ok ( $res, 'process_start' );
		$res = process_get_latest($job);
		is ( ref($res), 'HASH', 'process_get_latest' );
		# Le precedent s'est termine il y a 4 secondes, donc, >3 et <6.
		ok ( time() - $res->{'end'} > 3, 'process_get_latest' );
		ok ( time() - $res->{'end'} < 6, 'process_get_latest' );
		$res = process_stop($job);
		ok ( $res, 'process_stop' );
		BEGIN { $nb_sub2 += 5; }
	} else {
		sleep(1); # L'autre a deja fini de faire le boulot (hope so!)
		$res = process_start($job);
		ok ( $res, 'process_start' );
		$res = process_get_latest($job);
		is ( ref($res), 'HASH', 'process_get_latest' );
		ok ( time() - $res->{'end'} < 3, 'process_get_latest' );
		$res = process_abort($job);
		ok ( $res, 'process_abort' );
		BEGIN { $nb_sub2 += 4; }
		# Je suis le fils, je meurs.
		if ( ! Test::More->builder->is_passing() ) {
			exit(1);
		} else {
			exit(0);
		}
	}
	my $wait = wait();
	Test::More->builder->current_test($nb_sub2 - 1);
	is ( $?, 0, 'fin_du_fils' );
	BEGIN { $nb_sub2 += 1; }
};
BEGIN { $nbtest += 1; }





# Now, check we can do the same in 'nolock' fashion

setup_remember('./t','nolock');
is ( $FDN::Common::Remember::REP, './t', 'setup_remember' );
is ( $FDN::Common::Remember::MODE, 'nolock', 'setup_remember' );
BEGIN { $nbtest += 2; }

$job = 'rem2';

$res = process_start($job);
ok ( $res, 'process_start' );
ok ( -e "./t/$job.nolock", 'process_start' );
ok ( ! -e "./t/$job.lock", 'process_start' );
ok ( ! -e "./t/$job.latest", 'process_start' );
ok ( ! -e "./t/$job.data ", 'process_start' );
BEGIN { $nbtest += 5; }

$res = process_get_latest($job);
is ( $res, undef, 'process_get_latest' );
BEGIN { $nbtest += 1; }

$res = process_stop($job);
ok ( $res, 'process_stop' );
BEGIN { $nbtest += 1; }

my $nb_sub3;
subtest 'concurrent 3' => sub {
	plan tests => $nb_sub3;
	Test::More->builder->use_numbers(0);
	if ( fork() ) {
		$res = process_start($job);
		ok ( $res, 'process_start' );
		$res = process_get_latest($job);
		is ( ref($res), 'HASH', 'process_get_latest' );
		ok ( defined $res->{'start'}, 'process_get_latest' );
		ok ( defined $res->{'end'} , 'process_get_latest' );
		ok ( defined $res->{'hostname'}, 'process_get_latest' );
		ok ( defined $res->{'pid'}, 'process_get_latest' );
		ok ( time() - $res->{'start'} < 2, 'process_get_latest' );
		ok ( $res->{'pid'} == $$, 'process_get_latest' );
		BEGIN { $nb_sub3 += 8; };
		sleep(2);
		$res = process_stop($job);
		ok ( $res, 'process_stop' );
		BEGIN { $nb_sub3 += 1; }
    } else {
		sleep(1);
		$res = process_start($job);
		ok ( ! $res, 'process_start');
		$res = process_get_latest($job);
		is ( ref($res), 'HASH', 'process_get_latest');
		eval { $res = process_abort($job); };
		ok ( $@, 'process_abort' );
		BEGIN { $nb_sub3 += 3; }
		# Je suis le fils, je meurs.
		if ( ! Test::More->builder->is_passing() ) {
			exit(1); # Echec
		} else {
			exit(0); # Success
		}
	}
	my $wait = wait();
	Test::More->builder->current_test($nb_sub3 - 1);
	is ( $?, 0, 'fin_du_fils' );
	BEGIN { $nb_sub3 += 1; }
};
BEGIN { $nbtest += 1; }

sleep(4);

my $nb_sub4;
subtest 'concurrent 4' => sub {
	plan tests => $nb_sub4;
	Test::More->builder->use_numbers(0);
	if ( fork() ) {
		$res = process_start($job);
		ok ( $res, 'process_start' );
		$res = process_get_latest($job);
		is ( ref($res), 'HASH', 'process_get_latest' );
		# Le precedent s'est termine il y a 4 secondes, donc, >3 et <6.
		ok ( time() - $res->{'end'} > 3, 'process_get_latest' );
		ok ( time() - $res->{'end'} < 6, 'process_get_latest' );
		$res = process_stop($job);
		ok ( $res, 'process_stop' );
		BEGIN { $nb_sub4 += 5; }
	} else {
		sleep(1); # L'autre a deja fini de faire le boulot (hope so!)
		$res = process_start($job);
		ok ( $res, 'process_start' );
		$res = process_get_latest($job);
		is ( ref($res), 'HASH', 'process_get_latest' );
		ok ( time() - $res->{'end'} < 3, 'process_get_latest' );
		$res = process_abort($job);
		ok ( $res, 'process_abort' );
		BEGIN { $nb_sub4 += 4; }
		# Je suis le fils, je meurs.
		if ( ! Test::More->builder->is_passing() ) {
			exit(1);
		} else {
			exit(0);
		}
	}
	my $wait = wait();
	Test::More->builder->current_test($nb_sub4 - 1);
	is ( $?, 0, 'fin_du_fils' );
	BEGIN { $nb_sub4 += 1; }
};
BEGIN { $nbtest += 1; }

BEGIN { plan tests => $nbtest; }
